# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bf_nlu',
 'bf_nlu.cli',
 'bf_nlu.cli.arguments',
 'bf_nlu.cli.initial_project.actions',
 'bf_nlu.core',
 'bf_nlu.core.actions',
 'bf_nlu.core.brokers',
 'bf_nlu.core.channels',
 'bf_nlu.core.evaluation',
 'bf_nlu.core.featurizers',
 'bf_nlu.core.nlg',
 'bf_nlu.core.policies',
 'bf_nlu.core.training',
 'bf_nlu.core.training.converters',
 'bf_nlu.engine',
 'bf_nlu.engine.recipes',
 'bf_nlu.engine.runner',
 'bf_nlu.engine.storage',
 'bf_nlu.engine.training',
 'bf_nlu.graph_components',
 'bf_nlu.graph_components.converters',
 'bf_nlu.graph_components.providers',
 'bf_nlu.graph_components.validators',
 'bf_nlu.nlu',
 'bf_nlu.nlu.classifiers',
 'bf_nlu.nlu.emulators',
 'bf_nlu.nlu.extractors',
 'bf_nlu.nlu.featurizers',
 'bf_nlu.nlu.featurizers.dense_featurizer',
 'bf_nlu.nlu.featurizers.sparse_featurizer',
 'bf_nlu.nlu.selectors',
 'bf_nlu.nlu.tokenizers',
 'bf_nlu.nlu.utils',
 'bf_nlu.nlu.utils.hugging_face',
 'bf_nlu.shared',
 'bf_nlu.shared.core',
 'bf_nlu.shared.core.training_data',
 'bf_nlu.shared.core.training_data.story_reader',
 'bf_nlu.shared.core.training_data.story_writer',
 'bf_nlu.shared.importers',
 'bf_nlu.shared.nlu',
 'bf_nlu.shared.nlu.training_data',
 'bf_nlu.shared.nlu.training_data.formats',
 'bf_nlu.shared.nlu.training_data.schemas',
 'bf_nlu.shared.utils',
 'bf_nlu.shared.utils.schemas',
 'bf_nlu.utils',
 'bf_nlu.utils.tensorflow']

package_data = \
{'': ['*'],
 'bf_nlu.cli': ['initial_project/*',
                'initial_project/data/*',
                'initial_project/tests/*'],
 'bf_nlu.engine.recipes': ['config_files/*']}

install_requires = \
['CacheControl>=0.12.9,<0.13.0',
 'PyJWT[crypto]>=2.0.0,<3.0.0',
 'SQLAlchemy>=1.4.0,<1.5.0',
 'absl-py>=0.9,<0.14',
 'aio-pika>=6.7.1,<7.0.0',
 'aiohttp>=3.6,<3.8,!=3.7.4.post0',
 'apscheduler>=3.6,<3.8',
 'arabic_reshaper>=2.1.3,<3.0.0',
 'attrs>=19.3,<21.3',
 'boto3>=1.12,<2.0',
 'cloudpickle>=1.2,<1.7',
 'colorclass>=2.2,<2.3',
 'coloredlogs>=10,<16',
 'colorhash>=1.0.2,<1.1.0',
 'dask==2021.11.2',
 'datefinder>=0.7.1,<0.8.0',
 'fbmessenger>=6.0.0,<6.1.0',
 'google-auth<2',
 'jellyfish>=0.9.0,<0.10.0',
 'joblib>=0.15.1,<1.1.0',
 'jsonpickle>=1.3,<2.1',
 'jsonschema>=3.2,<4.5',
 'kafka-python>=1.4,<3.0',
 'matplotlib>=3.1,<3.4',
 'mattermostwrapper>=2.2,<2.3',
 'networkx>=2.4,<2.7',
 'numpy>=1.19.2,<1.20.0',
 'packaging>=20.0,<21.0',
 'prompt-toolkit>=2.0,<3.0',
 'psycopg2-binary>=2.8.2,<2.10.0',
 'pyTelegramBotAPI>=3.7.3,<4.0.0',
 'pyarabic>=0.6.14,<0.7.0',
 'pydot>=1.4,<1.5',
 'pykwalify>=1.7,<1.9',
 'pymongo[tls,srv]>=3.8,<3.11',
 'python-dateutil>=2.8,<2.9',
 'python-engineio>=4,<6,!=5.0.0',
 'python-socketio>=4.4,<6',
 'pytz>=2019.1,<2022.0',
 'questionary>=1.5.1,<1.11.0',
 'randomname>=0.1.5,<0.2.0',
 'rasa-sdk>=3.1.1,<3.2.0',
 'redis>=3.4,<4.0',
 'regex>=2020.6,<2022.5',
 'requests>=2.23,<3.0',
 'rocketchat_API>=0.6.31,<1.26.0',
 'ruamel.yaml>=0.16.5,<0.17.0',
 'sanic-cors>=2.0.0,<3.0.0',
 'sanic-jwt>=1.6.0,<2.0.0',
 'sanic-routing>=0.7.2,<0.8.0',
 'sanic>=21.12,<21.13',
 'scikit-learn>=0.22,<0.25',
 'scipy>=1.4.1,<2.0.0',
 'sentry-sdk>=0.17.0,<1.4.0',
 'setuptools>=41.0.0',
 'sklearn-crfsuite>=0.3,<0.4',
 'slackclient>=2.0.0,<3.0.0',
 'tarsafe>=0.0.3,<0.0.4',
 'tensorflow-addons>=0.15.0,<0.16.0',
 'tensorflow>=2.7.0,<2.8.0',
 'tensorflow_hub>=0.12.0,<0.13.0',
 'terminaltables>=3.1.0,<3.2.0',
 'tqdm>=4.31,<5.0',
 'twilio>=6.26,<6.51',
 'typing-extensions>=3.7.4,<4.0.0',
 'typing-utils>=0.1.0,<0.2.0',
 'ujson>=1.35,<6.0',
 'webexteamssdk>=1.1.1,<1.7.0']

extras_require = \
{':sys_platform != "win32"': ['tensorflow-text>=2.7.0,<3.0.0'],
 ':sys_platform == "win32"': ['colorama>=0.4.4,<0.5.0'],
 'full': ['spacy>=3.1,<4.0',
          'transformers>=4.13.0,<4.14.0',
          'sentencepiece[sentencepiece]>=0.1.96,<0.2.0',
          'jieba>=0.39,<0.43'],
 'gh-release-notes': ['github3.py>=1.3.0,<1.4.0'],
 'jieba': ['jieba>=0.39,<0.43'],
 'spacy': ['spacy>=3.1,<4.0'],
 'transformers': ['transformers>=4.13.0,<4.14.0',
                  'sentencepiece[sentencepiece]>=0.1.96,<0.2.0']}

entry_points = \
{'console_scripts': ['bf_nlu = bf_nlu.__main__:main']}

setup_kwargs = {
    'name': 'bf-nlu',
    'version': '1.1',
    'description': '',
    'long_description': '<h1 align="center">Bf Nlu BANKI Package</h1>\n\n',
    'author': 'botsfactory',
    'author_email': None,
    'maintainer': 'botsfactory',
    'maintainer_email': None,
    'url': '',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
